import re
import os
import ssl
import json
import time
import urllib2
import httplib
import logging
import threading
import mimetools
import mimetypes
import itertools

from functools import wraps
from types import BuiltinFunctionType
from __builtin__ import object as py_object

DEV_ENV = False  # Set True for development env

DEBUG = globals().get("DEBUG", False)
LOGGER_LEVEL = "DEBUG" if DEBUG else "INFO"


def _parse_version_from_params():
    version_ = globals().get("VERSION", {})

    title = version_.get("title")
    version = version_.get("version")
    company = re.search(r"<company>(.+)</company>", __doc__ or "")

    def _raise_err():
        raise ValueError("""Set constant VERSION = {"title": "YOUR SCRIPT TITLE", "version": 1.0}""")

    if title is None:
        title = re.search(r"<title>(.+)</title>", __doc__ or "")
        if title is None:
            _raise_err()
        else:
            version_["title"] = title.group(1)

    if version is None:
        version = re.search(r"<version>(.+)</version>", __doc__ or "")
        if version is None:
            _raise_err()
        else:
            version_["version"] = version.group(1)

    if company:
        version_["company"] = company.group(1)

    return version_


VERSION = _parse_version_from_params()


class TBotError(Exception):
    """Base Trassir Bot Error

    This exception can be raised from other
    threads throw host.timeout function
    """

    _host_api = host

    def __init__(self, *args):
        super(TBotError, self).__init__(*args)
        self._host_api.timeout(1, self.rise_from_thread)

    def rise_from_thread(self):
        raise self


class ServerKeyError(TBotError):
    """Raises when script can't get server key"""


class TgUsersError(TBotError):
    """Raises with bad/empty telegram user id's"""


class TBotResponseError(TBotError):
    """Raises with bad/empty telegram user id's"""


class KeepAliveError(TBotError):
    """Raises if Bad keep alive error"""


class TBotStartingError(TBotError):
    """Raises if some problems on start"""


class HostLogHandler(logging.Handler):
    """Trassir main log handler"""

    def __init__(self, host_api=host):
        super(HostLogHandler, self).__init__()
        self._host_api = host_api

    def emit(self, record):
        msg = self.format(record)
        self._host_api.log_message(msg)


class PopupHandler(logging.Handler):
    """Trassir popup handler"""

    def __init__(self, host_api=host):
        super(PopupHandler, self).__init__()
        self._host_api = host_api
        self._popups = {
            "CRITICAL": host_api.error,
            "FATAL": host_api.error,
            "ERROR": host_api.error,
            "WARN": host_api.alert,
            "WARNING": host_api.alert,
            "INFO": host_api.message,
            "DEBUG": host_api.message,
            "NOTSET": host_api.message,
        }

    def emit(self, record):
        msg = self.format(record)
        self._popups[record.levelname](msg)


class DuplicateFilter(logging.Filter):
    """Suppressing multiple messages with same content.

    Tracking last logged record and filter out any
    repeated (similar) records. Output something more rsyslog style.

    Example:
        --- The last message repeated 3 times
    """

    def __init__(self):
        super(DuplicateFilter, self).__init__()
        self._last_log = None
        self._last_log_count = 1

    def filter(self, record):
        record.duplicates = ""
        current_log = (record.module, record.levelno, record.msg)
        if current_log == self._last_log:
            self._last_log_count += 1
            return False
        else:
            if self._last_log_count > 1:
                record.duplicates = (
                    "--- The last message repeated %s times\n" % self._last_log_count
                )
            self._last_log = current_log
            self._last_log_count = 1
            return True


def _get_logger(
    host_log="WARNING",
    popup_log="ERROR",
    file_log=None,
    file_name=None,
    file_max_bytes=5 * 1024 * 1024,
    file_backup_count=2,
):
    """Возвращает логгер с предустановленными хэндлерами

    Args:
        host_log (str, optional): Уровень логирования в основной лог.
        popup_log (str, optional): Уровень логирования во всплывающих сообщениях.
        file_log (str, optional): Уровень логирования в отдельный файл.
        file_name (str, optional): Имя файла для логирования.
        file_max_bytes (int, optional): Максимальный размер файла лога в байтах.
        file_backup_count (int, optional): Макссимальное кол-во бэкапов лога.

    Returns:
        logging.logger: Настроенный логгер
    """
    logger_ = logging.getLogger(__name__)
    logger_.setLevel("DEBUG")

    def _remove_handlers():
        """Close and remove handlers on disable script"""
        logger_.info("Remove handlers...")
        for handler in logger_.handlers[:]:
            handler.close()
            logger_.removeHandler(handler)

    try:
        host.register_finalizer(_remove_handlers)
        allow_file_handler = True
    except AttributeError:
        # Trassir ARM
        # AttributeError: 'module' object has no attribute 'register_finalizer'
        for handler in logger_.handlers[:]:
            handler.close()
            logger_.removeHandler(handler)
        allow_file_handler = False

    if host_log:
        host_handler = HostLogHandler()
        host_handler.setLevel(host_log)
        host_formatter = logging.Formatter(
            "[%(levelname)-8s] %(lineno)-4s <%(funcName)s> - %(message)s"
        )
        host_handler.setFormatter(host_formatter)
        logger_.addHandler(host_handler)

    if popup_log:
        popup_handler = PopupHandler()
        popup_handler.setLevel(popup_log)
        popup_formatter = logging.Formatter(
            fmt="<b>[%(levelname)s]</b> Line: %(lineno)s<br><i>%(message).630s</i>"
        )
        popup_handler.setFormatter(popup_formatter)
        logger_.addHandler(popup_handler)

    if file_log and allow_file_handler:
        try:
            from logging.handlers import RotatingFileHandler
        except ImportError:
            # Trassir ARM
            # ImportError: No module named cPickle
            pass
        else:
            if file_name is None:
                file_name = cls.get_script_name()

            if not file_name.endswith(".log"):
                file_name = "{}.log".format(file_name)

            file_path = os.path.join(cls.get_screenshot_folder(), file_name)
            file_path = cls.win_encode_path(file_path)

            file_handler = RotatingFileHandler(
                file_path, maxBytes=file_max_bytes, backupCount=file_backup_count
            )
            file_handler.setLevel(file_log)
            file_formatter = logging.Formatter(
                fmt="%(duplicates)s%(asctime)s [%(levelname)-8s] %(lineno)-4s <%(funcName)s> - %(message)s",
                datefmt="%Y/%m/%d %H:%M:%S",
            )
            file_handler.setFormatter(file_formatter)
            file_handler.addFilter(DuplicateFilter())
            logger_.addHandler(file_handler)

    return logger_


logger = globals().get("logger")
if logger is None:
    if DEBUG:
        logger = _get_logger(host_log="DEBUG", popup_log="WARNING", file_log="DEBUG")
    else:
        logger = _get_logger()


class Utils:
    def __init__(self):
        pass

    @staticmethod
    def catch_request_exceptions(func):
        """Catch request errors"""

        @wraps(func)
        def wrapped(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except urllib2.HTTPError as e:
                code = e.code
                res = e.read()
                logger.warning(
                    "HTTPError: kwargs: {kwargs}, code: {code}, res: {res}".format(
                        kwargs=kwargs, code=code, res=res
                    )
                )
                return code, res
            except urllib2.URLError as e:
                logger.warning(
                    "URLError: kwargs: {kwargs}, reason: {e.reason}".format(
                        kwargs=kwargs, e=e
                    )
                )
                return "URLError", e.reason
            except httplib.HTTPException as e:
                logger.warning(
                    "HTTPException: kwargs: {kwargs}, err: {e}".format(
                        kwargs=kwargs, e=e
                    )
                )
                return "HTTPException", e
            except ssl.SSLError as e:
                logger.warning(
                    "SSLError: kwargs: {kwargs}, errno: {e.errno}, err: {e}".format(
                        kwargs=kwargs, e=e
                    )
                )
                return "SSLError", "[{}]: {}".format(e.errno, e)

        return wrapped

    @staticmethod
    def run_as_thread(fn):
        """Run function as thread"""

        @wraps(fn)
        def run(*args, **kwargs):
            t = threading.Thread(target=fn, args=args, kwargs=kwargs)
            t.daemon = True
            t.start()
            return t

        return run

    @staticmethod
    def join_paths(*args):
        """Returns path from args

        Encode to cp1251 for windows server
        And return full path from args
        """
        full_path = os.path.join(*args)
        if os.name == "nt":
            full_path = full_path.decode("utf8").encode("cp1251")
        return full_path

    @staticmethod
    def do_nothing(*args, **kwargs):
        """Speaks by itself"""
        pass

    @staticmethod
    def save_json(filename, data):
        """Save json file

        Args:
            filename (str) : Full path to json file
            data (str | dict | list) : Data to save in file
        """
        if not isinstance(data, str):
            data = json.dumps(data)
        with open(filename, "w") as json_file:
            json_file.write(data)

    @staticmethod
    def load_json(filename):
        """Load json data from file

        Returns empty dict if loading failed

        Args:
            filename (str) : Full path to json file
        """
        data = {}
        if os.path.isfile(filename):
            with open(filename, "r") as json_file:
                try:
                    data = json.loads(json_file.read())
                except ValueError:
                    pass
        return data


class MultiPartForm(py_object):
    """Accumulate the data to be used when posting a form."""

    def __init__(self):
        self.form_fields = []
        self.files = []
        self.boundary = mimetools.choose_boundary()
        return

    def get_content_type(self):
        return "multipart/form-data; boundary=%s" % self.boundary

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
        return

    def add_file(self, field_name, filename, file_handle, mimetype=None):
        """Add a file to be uploaded."""
        body = file_handle
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or "application/octet-stream"
            logger.debug("mimetype: %s" % mimetype)

        if os.name == "nt":
            filename = filename.decode("cp1251")
        self.files.append((field_name, filename, mimetype, body))
        return

    def __str__(self):
        """Return a string representing the form data, including attached files."""
        # Build a list of lists, each containing "lines" of the
        # request.  Each part is separated by a boundary string.
        # Once the list is built, return a string where each
        # line is separated by '\r\n'.
        parts = []
        part_boundary = "--" + self.boundary

        # Add the form fields
        parts.extend(
            [
                part_boundary,
                'Content-Disposition: form-data; name="%s"' % name,
                "",
                value,
            ]
            for name, value in self.form_fields
        )

        # Add the files to upload
        parts.extend(
            [
                part_boundary,
                'Content-Disposition: file; name="%s"; filename="%s"'
                % (field_name, filename),
                "Content-Type: %s" % content_type,
                "",
                body,
            ]
            for field_name, filename, content_type, body in self.files
        )

        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append("--" + self.boundary + "--")
        flattened.append("")
        return "\r\n".join(str(fl) for fl in flattened)


class HTTPRequests(py_object):
    _REQUEST_TIMEOUT = 5  # Timeout

    def __init__(self, opener=None):
        """Class to work with http requests

        Args:
            opener (urllib2.OpenerDirector): Class to open URLs. Default `None`
                https://docs.python.org/2/library/urllib2.html#urllib2.build_opener
        """
        self._opener = opener or self._build_opener()

    @staticmethod
    def _build_opener():
        """Build default opener for requests"""
        try:
            ctx = ssl.create_default_context()
        except AttributeError as err:
            raise AttributeError(
                "Возникла ошибка при запуске скрипта<br><br>"
                "<span style='color: #000000;'>В 90% случаях помогает обновление Trassir до актуальной версии<br>"
                "Еще в 7% случаях помогает 'чистая' установка Trassir в отдельную папку<br>"
                "Очень редко ~2% случаев помогает обнвление Python (если он у вас установлен)<br>"
                "<br>"
                "Если Вам ничего не помогло - напишите нашим специалистам на почту "
                "<a href='mailto:support@dssl.ru'>support@dssl.ru</a><br>"
                "Приложите к письму скриншот с данной ошибкой, а также дампы сервера<br>"
                "</span><br>"
                ">>> {err}".format(err=err)
            )
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        handler_ = urllib2.HTTPSHandler(context=ctx)
        return urllib2.build_opener(handler_)

    @staticmethod
    def _encode_data(data):
        """Returns json dumps data"""
        assert isinstance(data, dict), "data must be dict, not {}".format(
            type(data).__name__
        )
        return json.dumps(data)

    @Utils.catch_request_exceptions
    def post_json(self, url, data=None):
        """Do post request

        Args:
            url (str) : Request url
            data (dict, optional): Request data. Default `None`
        """
        request = urllib2.Request(url)

        if data is None:
            data = {}

        request.add_data(self._encode_data(data))
        request.add_header("Content-Type", "application/json")

        response = self._opener.open(request, timeout=self._REQUEST_TIMEOUT)
        code, res = response.code, response.read()

        logger.debug(
            "../{} response [{}] - {}".format(url.rsplit("/", 1)[-1], code, res)
        )

        return code, res

    # noinspection PyIncorrectDocstring
    @Utils.catch_request_exceptions
    def post_file(self, url, **kwargs):
        """Do post form data request

        Args:
            url (str) : Request url
            data (Dict[], optional): Form data. Default `None`
            filename (str, optional): Custom file name. Default `None`

            photo (str, optional): Image filepath. Used to send image.
            document (str, optional): Doc filepath. Used to send document.
        """
        request = urllib2.Request(url)
        form = MultiPartForm()

        data = kwargs.pop("data", None)
        if data is None:
            data = {}
        for key, value in data.iteritems():
            form.add_field(key, value)

        filename = kwargs.pop("filename", None)
        for key, filepath in kwargs.iteritems():
            if filename is None:
                filename = os.path.basename(filepath)
            logger.debug("filename: %s" % filename)
            form.add_file(key, filename, open(filepath, "rb").read())

        try:
            body = str(form)
        except UnicodeDecodeError:
            logger.exception("UnicodeDecodeError")
            return

        request.add_header("User-agent", "tbot v{}".format(VERSION["version"]))
        request.add_header("Content-type", form.get_content_type())
        request.add_header("Content-length", len(body))
        request.add_data(body)

        response = self._opener.open(request, timeout=self._REQUEST_TIMEOUT)
        code, res = response.code, response.read()

        logger.debug(
            "../{} response [{}] - {}".format(url.rsplit("/", 1)[-1], code, res)
        )

        return code, res


# noinspection PyBroadException
class TBotAPI(HTTPRequests):
    """Api for TBot service"""

    if DEV_ENV:
        _BASE_URL = "https://172.20.0.9:8443/api/1.0"
    else:
        _BASE_URL = "https://bot.trassir.com:443/api/1.0"

    _SERVICE_VERSION = 0.64

    _AWAITING_IMAGE = 10  # sec
    _KEEP_ALIVE_TIMEOUT = 60 * 60 * 4  # sec

    _ALREADY_STARTED = False

    def __init__(self, *args, **kwargs):
        """Api for @trassirbot

        Args:
            opener (urllib2.OpenerDirector): Class to open URLs. Default `None`
                https://docs.python.org/2/library/urllib2.html#urllib2.build_opener
            tg_users (Dict[admin: List[int], operator: List[int]]): Telegram users. Default: `None`
            is_bot (bool): True if script is bot. Default: `False`
            host_api (module, optional) : Trassir host module. Default: `host`
        """
        if args:
            raise TypeError("TBotAPI() Takes Keyword-Only Arguments!")

        opener = kwargs.get("opener", None)
        tg_users = kwargs.get("tg_users", None)
        is_bot = kwargs.get("is_bot", False)
        host_api = kwargs.get("host_api", host)

        super(TBotAPI, self).__init__(opener)

        if host_api.settings("").guid == "client":
            raise EnvironmentError("Скрипт не доступен для Trassir Client")

        logger.info(
            "TBotAPI(tg_users={tg_users}, is_bot={is_bot}, opener={opener})".format(
                tg_users=repr(tg_users), is_bot=repr(is_bot), opener=repr(opener)
            )
        )

        if TBotAPI._ALREADY_STARTED:
            raise TBotError("You can init TBotAPI only once!")
        TBotAPI._ALREADY_STARTED = True

        if tg_users is not None:
            if not isinstance(tg_users, dict):
                raise TBotError(
                    "Expected dict, got {} with argument: tg_users={}".format(
                        type(tg_users).__name__, repr(tg_users)
                    )
                )
        if not isinstance(is_bot, bool):
            raise TBotError(
                "Expected bool, got {} with argument: is_bot={}".format(
                    type(is_bot).__name__, repr(is_bot)
                )
            )
        if tg_users is None and is_bot:
            raise TBotError("TBot can't start without users")

        self.is_arm = False
        self._host_api = host_api
        self._tg_users = tg_users
        self._is_bot = is_bot
        self._script, self._server = None, None

        self._greeting()

        if is_bot is False:
            self._keep_alive()

    def _get_server_key(self):
        """Return guardant usb key or online activated SRL

        Raises:
            ServerKeyError: If server key is empty (usually on client)
        """
        if not isinstance(self._host_api.license_id, BuiltinFunctionType):
            raise EnvironmentError("Broken license_id function!")

        try:
            license_info_str = self._host_api.settings("health")["license_info"]
            license_info = json.loads(license_info_str)
        except (KeyError, ValueError):
            # KeyError if old server
            # ValueError if license_info is empty
            license_info = {}

        server_key = license_info.get("id", self._host_api.license_id())
        if server_key == "0x00000001":
            # Trassir ARM
            self.is_arm = True
            srl = None
            with open("/data/ Trassir 3 License.txt", "r") as license:
                for line in license:
                    if line.startswith("SRL:"):
                        srl = re.search(r"[\w\d]{4}-[\w\d]{4}-[\w\d]{4}-[\w\d]{4}", line)
                        break
            if srl is None:
                raise ServerKeyError("Can't load SRL key")
            else:
                server_key = srl.group()

        if not server_key:
            raise ServerKeyError("Server key is not available")

        if server_key.startswith("0x"):
            server_key = server_key[2:]

        return server_key

    @staticmethod
    def prepare_users(tg_users):
        """Check and prepare telegram users

        Args:
            tg_users (List[int] | List[str] | str | int): Telegram id list.
                 Prefer to use the List[int].

        Returns:
            List[int]: Telegram users id list

        Raises:
            TgUsersError: if can't make int from user id
        """
        if not tg_users:
            raise TgUsersError("No telegram users id")

        if isinstance(tg_users, str):
            if "," in tg_users:
                tg_users = tg_users.split(",")

        if not isinstance(tg_users, list):
            tg_users = [tg_users]

        tg_users_ = set()
        for tg_user in tg_users:
            if not tg_user:
                continue
            try:
                tg_users_.add(int(tg_user))
            except ValueError:
                raise TgUsersError("Bad telegram user id: {}".format(tg_user))

        if not tg_users_:
            raise TgUsersError("No telegram users id")

        return list(tg_users_)

    def _check_response(self, code, res, request="Undefined"):
        """Check response
            400
                1 - No users to send
                2 - No content
                3 - No callback query id

                101 - Broken telegram user id
                102 - User can't be admin and operator
                103 - No users for bot

                111 - Broken request data

            401
                1 - No server key
                2 - No script guid
                3 - Unknown script

            403
                1 - Server is banned
                2 - Script is not bot

            413
                1 - File to large

            415
                1 - Not available file extension

        Raises:
            TBotError: if some errors
            KeepAliveError: if script not found on server
        """
        try:
            res = json.loads(str(res))
        except ValueError:
            logger.warning("Request({}) error: [{}] {}".format(request, code, res))
            return code, res

        error_code = res.get("error_code", 999)
        error_message = res.get("message", res)
        if code == 400:
            if error_code > 100:
                raise TBotError(error_message)

        elif code == 401:
            error_code = res.get("error_code", 0)
            if self._is_bot and error_code == 3:
                """Script was removed from database because 
                another tbot script was started"""

                raise KeepAliveError(
                    "Your script did't send keep alive message for a long time\n "
                    "or you start another TBot script on this server"
                )

            else:
                raise TBotError(error_message)

        elif code == 403:
            raise TBotError(error_message)

        return error_code, error_message

    def _greeting(self):
        """Send greeting request from script

        Raises:
            TBotStartingError: if get error from server
        """
        script = self._host_api.stats().parent()
        server = self._host_api.settings("")
        server.key = self._get_server_key()

        # try:
        #     version_ = float(VERSION["VERSION"])
        # except ValueError:
        #     raise TBotError("Can't parse script version to float")

        data = {
            "server": {"key": server.key, "guid": server.guid, "name": server.name},
            "script": {
                "guid": script.guid,
                "name": script.name,
                "version": {
                    "title": VERSION["title"],
                    "version": VERSION["version"],
                    "service_version": self._SERVICE_VERSION,
                },
                "is_bot": self._is_bot,
            },
            "users": self._tg_users,
        }

        url = "{}/startScript".format(self._BASE_URL)
        code, res = self.post_json(url, data)
        if code != 200:
            error_code, error_message = self._check_response(
                code, res, request="startScript"
            )
            if os.name == "nt":
                error_code = str(error_code).decode("cp1251")
                error_message = str(error_message).decode("cp1251")
            raise TBotStartingError("{}".format(error_message or error_code))

        self._script = script
        self._server = server

    def _keep_alive(self):
        """Send keep alive command to server

        Note:
            Not need for TBot, because it send data every 5 sec
        """
        data = {"server_key": self._server.key, "script_guid": self._script.guid}
        url = "{}/keepAliveScript".format(self._BASE_URL)
        response = self.post_json(url, data)
        self._check_response(*response, request="keepAlive")

        self._host_api.timeout(1000 * self._KEEP_ALIVE_TIMEOUT, self._keep_alive)

    def send_server_data(self, server_data):
        """Send server data

        Args:
            server_data (Dict[]) : Server data
        """
        data = {
            "server_key": self._server.key,
            "script_guid": self._script.guid,
            "data": server_data,
        }
        url = "{}/updateServerData".format(self._BASE_URL)
        response = self.post_json(url, data)
        self._check_response(*response, request="updateServerData")

        return response

    @Utils.run_as_thread
    def send_chat_action(self, tg_users, action):
        """Send telegram action

        Note:
            https://core.telegram.org/bots/api#sendchataction

        Args:
            tg_users (List[int] | List[str] | str | int): Telegram users id list.
            action (str): Type of action to broadcast.

        Raises:
            TgUsersError: if can't make int from user id
        """
        tg_users = self.prepare_users(tg_users)
        try:
            data = {
                "chat_ids": tg_users,
                "action": action,
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }
            url = "{}/sendChatAction".format(self._BASE_URL)
            response = self.post_json(url, data)

        except Exception as err:
            logger.exception("send_chat_action error", exc_info=True)
            raise TBotError(str(err))

        self._check_response(*response, request="sendChatAction")
        return response

    @Utils.run_as_thread
    def callback_answer(self, callback_query_id, message, show_alert=False):
        """Send callback answer

        Args:
            callback_query_id (str): Unique identifier for the query to be answered
            message (str) : Message to sending
            show_alert (bool, optional): If true, an alert will be shown by the client
                instead of a notification at the top of the chat screen. Default `False`
        """
        try:
            data = {
                "text": str(message),
                "callback_query_id": callback_query_id,
                "show_alert": show_alert,
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }

            url = "{}/answerCallbackQuery".format(self._BASE_URL)
            response = self.post_json(url, data)

        except Exception as err:
            logger.exception("callback_answer error", exc_info=True)
            raise TBotError(str(err))

        self._check_response(*response, request="answerCallbackQuery")
        return response

    @Utils.run_as_thread
    def send_message(self, tg_users, message):
        """Send text message

        You can use HTML style tags in message.
        Supported tags: b, i, a, code, pre. For more information
        go to https://core.telegram.org/bots/api#html-style

        Warnings:
            You get error message in telegram if your text contains invalid tags
            Example: Error sending message. Can't parse entities: unsupported start tag...

        Args:
            tg_users (List[int] | List[str] | str | int): Telegram users id list.
            message (str): Text message to send.

        Raises:
            TgUsersError: if can't make int from user id or user id is empty
        """
        tg_users = self.prepare_users(tg_users)
        try:
            data = {
                "chat_ids": tg_users,
                "text": str(message),
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }

            url = "{}/sendMessage".format(self._BASE_URL)
            response = self.post_json(url, data)

        except Exception as err:
            logger.exception("send_message error", exc_info=True)
            raise TBotError(str(err))

        self._check_response(*response, request="sendMessage")
        return response

    def check_file(self, full_path):
        """Check if file exists, run only in thread!

        Args:
            full_path (str) : Full path to file
        """
        for _ in xrange(self._AWAITING_IMAGE):
            if os.path.isfile(full_path):
                return True
            time.sleep(1)
        return False

    @Utils.run_as_thread
    def send_image(self, tg_users, image_path, caption="", remove=False):
        """Send image

        You can use HTML style tags in caption.
        Supported tags: b, i, a, code, pre. For more information
        go to https://core.telegram.org/bots/api#html-style

        Warnings:
            You get error message in telegram if your caption contains invalid tags
            Example: Error sending image. Can't parse entities: unsupported start tag...

        Note:
            Working in thread to await screenshot creating
            Making 10 tries (self._AWAITING_IMAGE = 10) every second

        Args:
            tg_users (List[int] | List[str] | str | int): Telegram users id list.
            image_path (str): Full path to image
            caption (str, optional): Caption to sending image. Default `""`
            remove (bool, optional): Remove file after sending or not. Default `False`

        Raises:
            TgUsersError: if can't make int from user id or user id is empty
        """
        logger.debug("tg_users: %s" % tg_users)
        tg_users = self.prepare_users(tg_users)
        try:
            self.send_chat_action(tg_users, "upload_photo")
            data = {
                "chat_ids": ",".join(str(tg_id) for tg_id in tg_users),
                "caption": caption,
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }
            if self.check_file(image_path):
                url = "{}/sendPhoto".format(self._BASE_URL)
                response = self.post_file(
                    url, photo=image_path, filename=None, data=data
                )
                if remove:
                    os.remove(image_path)
            else:
                response = self.send_message(
                    tg_users, "No file {}".format(caption if caption else image_path)
                )
        except Exception as err:
            logger.exception("send_image error")
            raise TBotError(str(err))

        self._check_response(*response, request="sendPhoto")
        return response

    @Utils.run_as_thread
    def send_image_album(self, tg_users, image_paths, captions=None, remove=False):
        """Send images in album

        You can use HTML style tags in caption.
        Supported tags: b, i, a, code, pre. For more information
        go to https://core.telegram.org/bots/api#html-style

        Warnings:
            You get error message in telegram if your caption contains invalid tags
            Example: Error sending image. Can't parse entities: unsupported start tag...

        Note:
            Working in thread to await screenshot creating
            Making 10 tries (self._AWAITING_IMAGE = 10) every second



        Args:
            tg_users (List[int] | List[str] | str | int): Telegram users id list.
            image_paths (List[str]): List of full path to each image
            captions (List[str], optional): List of caption to each image. Default `None`
            remove (bool, optional): Remove file after sending or not. Default `False`

        Raises:
            TgUsersError: if can't make int from user id or user id is empty
        """
        logger.debug("tg_users: %s" % tg_users)
        tg_users = self.prepare_users(tg_users)
        try:
            # self.send_chat_action(tg_users, "upload_photo")
            data = {
                "chat_ids": ",".join(str(tg_id) for tg_id in tg_users),
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }
            if not isinstance(image_paths, list):
                raise TBotError(
                    "Expected list for image_paths, got {}".format(
                        type(image_paths).__name__
                    )
                )

            if captions is None:
                captions = []

            while len(captions) < len(image_paths):
                captions.append(None)

            kwargs = {}
            remove_list = []
            for idx, image in enumerate(image_paths):
                if self.check_file(image):
                    kwargs["photo_%s" % idx] = image
                    remove_list.append(image)
                    if captions[idx] is not None:
                        data["photo_%s_caption" % idx] = captions[idx]

                else:
                    self.send_message(
                        tg_users, "No image {}".format(captions[idx] or image)
                    )

            kwargs["data"] = data
            url = "{}/sendPhotoAlbum".format(self._BASE_URL)
            response = self.post_file(url, **kwargs)

            if remove:
                for image in remove_list:
                    os.remove(image)

        except Exception as err:
            logger.exception("send_image error")
            raise TBotError(str(err))

        self._check_response(*response, request="sendPhoto")
        return response

    @Utils.run_as_thread
    def send_document(
        self, tg_users, doc_path, filename=None, caption="", remove=False
    ):
        """Send document to telegram

        You can use HTML style tags in caption.
        Supported tags: b, i, a, code, pre. For more information
        go to https://core.telegram.org/bots/api#html-style

        Warnings:
            You get error message in telegram if your caption contains invalid tags
            Example: Error sending document. Can't parse entities: unsupported start tag...

        Note:
            Working in thread to await file creating
            Making 10 tries (self._AWAITING_IMAGE = 10) every second

        Args:
            tg_users (List[int] | List[str] | str | int): Telegram users id list.
            doc_path (str): Full path to document
            filename (str, optional): Custom filename. Default `None`.
            caption (str, optional): Caption to sending document. Default `""`.
            remove (bool, optional): Remove file after sending or not. Default `False`

        Raises:
            TgUsersError: if can't make int from user id or user id is empty
        """
        tg_users = self.prepare_users(tg_users)
        try:
            self.send_chat_action(tg_users, "upload_document")

            data = {
                "chat_ids": ",".join(str(tg_id) for tg_id in tg_users),
                "caption": caption,
                "server_key": self._server.key,
                "script_guid": self._script.guid,
            }

            if self.check_file(doc_path):
                url = "{}/sendDocument".format(self._BASE_URL)
                response = self.post_file(
                    url, document=doc_path, filename=filename, data=data
                )
                if remove:
                    os.remove(doc_path)
            else:
                response = self.send_message(
                    tg_users, "No file {}".format(caption if caption else doc_path)
                )

        except Exception as err:
            logger.exception("send_document error", exc_info=True)
            raise TBotError(str(err))

        self._check_response(*response, request="sendDocument")
        return response